/* -*-c-*-
 *
 * $Id: blowfish.c,v 1.1 1999/09/03 08:41:11 mdw Exp $
 *
 * The Blowfish block cipher
 *
 * (c) 1998 Straylight/Edgeware
 */

/*----- Licensing notice --------------------------------------------------* 
 *
 * This file is part of Catacomb.
 *
 * Catacomb is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 * 
 * Catacomb is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with Catacomb; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 */

/*----- Revision history --------------------------------------------------* 
 *
 * $Log: blowfish.c,v $
 * Revision 1.1  1999/09/03 08:41:11  mdw
 * Initial import.
 *
 */

/*----- Header files ------------------------------------------------------*/

#include <mLib/bits.h>

#include "blowfish.h"
#include "bf_ikey.h"
#include "blkc.h"
#include "paranoia.h"

/*----- Global variables --------------------------------------------------*/

static blowfish_ctx ikey = BLOWFISH_IKEY;

/*----- Macros ------------------------------------------------------------*/

#define ROUND(k, x, y, r)						\
  ((x) ^= (k)->p[(r)],							\
   (y) ^= (((((k)->s0[((x) >> 24) & MASK8]) +				\
	     ((k)->s1[((x) >> 16) & MASK8])) ^				\
	    ((k)->s2[((x) >>  8) & MASK8])) +				\
	   ((k)->s3[((x) >>  0) & MASK8])))

#define EBLK(k, a, b, c, d) do {					\
  uint32 _x = (a);							\
  uint32 _y = (b);							\
  ROUND((k), _x, _y,  0);						\
  ROUND((k), _y, _x,  1);						\
  ROUND((k), _x, _y,  2);						\
  ROUND((k), _y, _x,  3);						\
  ROUND((k), _x, _y,  4);						\
  ROUND((k), _y, _x,  5);						\
  ROUND((k), _x, _y,  6);						\
  ROUND((k), _y, _x,  7);						\
  ROUND((k), _x, _y,  8);						\
  ROUND((k), _y, _x,  9);						\
  ROUND((k), _x, _y, 10);						\
  ROUND((k), _y, _x, 11);						\
  ROUND((k), _x, _y, 12);						\
  ROUND((k), _y, _x, 13);						\
  ROUND((k), _x, _y, 14);						\
  ROUND((k), _y, _x, 15);						\
  (c) = _y ^ (k)->p[17];						\
  (d) = _x ^ (k)->p[16];						\
} while (0)

#define DBLK(k, a, b, c, d) do {					\
  uint32 _x = (a);							\
  uint32 _y = (b);							\
  ROUND((k), _x, _y, 17);						\
  ROUND((k), _y, _x, 16);						\
  ROUND((k), _x, _y, 15);						\
  ROUND((k), _y, _x, 14);						\
  ROUND((k), _x, _y, 13);						\
  ROUND((k), _y, _x, 12);						\
  ROUND((k), _x, _y, 11);						\
  ROUND((k), _y, _x, 10);						\
  ROUND((k), _x, _y,  9);						\
  ROUND((k), _y, _x,  8);						\
  ROUND((k), _x, _y,  7);						\
  ROUND((k), _y, _x,  6);						\
  ROUND((k), _x, _y,  5);						\
  ROUND((k), _y, _x,  4);						\
  ROUND((k), _x, _y,  3);						\
  ROUND((k), _y, _x,  2);						\
  (c) = _y ^ (k)->p[0];							\
  (d) = _x ^ (k)->p[1];							\
} while (0)

/*----- Low-level encryption interface ------------------------------------*/

/* --- @blowfish_init@ --- *
 *
 * Arguments:	@blowfish_ctx *k@ = pointer to key block to fill in
 *		@const void *buf@ = pointer to buffer of key material
 *		@size_t sz@ = size of key material
 *
 * Returns:	---
 *
 * Use:		Initializes a Blowfish key buffer.  Blowfish accepts
 *		a more-or-less arbitrary size key.
 */

void blowfish_init(blowfish_ctx *k, const void *buf, size_t sz)
{
  /* --- Copy the initial value over --- */

  memcpy(k, &ikey, sizeof(ikey));

  /* --- Initialize the %$P$% array --- */

  {
    const octet *p = buf;
    const octet *q = p + sz;
    int i = 0, j = 0;
    uint32 x = 0;

    while (i < 18) {
      x = (x << 8) | U8(*p++);
      if (p >= q)
	p = buf;
      if (++j >= 4) {
	k->p[i++] ^= x;
	x = 0;
	j = 0;
      }
    }

    x = 0;
  }

  /* --- Now mangle the complete array of keys --- */

  {
    uint32 b[2];
    int i;

    b[0] = b[1] = 0;

    for (i = 0; i < 18; i += 2) {
      blowfish_eblk(k, b, b);
      k->p[i] = b[0]; k->p[i + 1] = b[1];
    }

    for (i = 0; i < 256; i += 2) {
      blowfish_eblk(k, b, b);
      k->s0[i] = b[0]; k->s0[i + 1] = b[1];
    }

    for (i = 0; i < 256; i += 2) {
      blowfish_eblk(k, b, b);
      k->s1[i] = b[0]; k->s1[i + 1] = b[1];
    }

    for (i = 0; i < 256; i += 2) {
      blowfish_eblk(k, b, b);
      k->s2[i] = b[0]; k->s2[i + 1] = b[1];
    }

    for (i = 0; i < 256; i += 2) {
      blowfish_eblk(k, b, b);
      k->s3[i] = b[0]; k->s3[i + 1] = b[1];
    }

    BURN(b);
  }
}

/* --- @blowfish_eblk@, @blowfish_dblk@ --- *
 *
 * Arguments:	@const blowfish_ctx *k@ = pointer to key block
 *		@const uint32 s[2]@ = pointer to source block
 *		@uint32 d[2]@ = pointer to destination block
 *
 * Returns:	---
 *
 * Use:		Low-level block encryption and decryption.
 */

void blowfish_eblk(const blowfish_ctx *k, const uint32 *s, uint32 *d)
{
  EBLK(k, s[0], s[1], d[0], d[1]);
}

void blowfish_dblk(const blowfish_ctx *k, const uint32 *s, uint32 *d)
{
  DBLK(k, s[0], s[1], d[0], d[1]);
}

BLKC_TEST(BLOWFISH, blowfish)

/*----- That's all, folks -------------------------------------------------*/
